import time
import typing
import mimetypes

from apistar.http import Response, StrMapping, StrPairs

from ..helper import parse_date, parse_range_header, file_iter_range


class FileResponse(Response):
    def __init__(self,
                 content: typing.Any,
                 filename: str=None,
                 status_code: int=200,
                 headers: typing.Union[StrMapping, StrPairs]=None,
                 exc_info=None,
                 media_type: str="auto",
                 charset: str='utf-8',
                 download: bool=True,
                 last_modified: str=None,
                 if_modified_since: str=None,
                 ranges: str=None,
    ) -> None:
        self.filename = filename
        self.media_type = media_type or self.media_type
        self.charset = charset
        self.download = download
        self.last_modified = last_modified
        self.if_modified_since = if_modified_since
        self.ranges = ranges
        super(FileResponse, self).__init__(content, status_code, headers, exc_info)

    def render(self, content: typing.Any) -> bytes:
        if hasattr(content, "read"):
            content = content.read()
            if not self.filename:
                self.filename = getattr(content, "filename", None)
        if isinstance(content, bytes):
            return content
        elif isinstance(content, str) and self.charset is not None:
            return content.encode(self.charset)

        valid_types = "bytes" if self.charset is None else "string or bytes"
        raise RuntimeError(
            "%s content must be %s. Got %s." %
            (self.__class__.__name__, valid_types, type(content).__name__)
        )

    def set_default_headers(self):
        if 'Content-Length' not in self.headers:
            self.headers['Content-Length'] = str(len(self.content))

        assert self.filename, "Filename must specific."

        if self.media_type == 'auto':
            self.media_type, encoding = mimetypes.guess_type(self.filename)
            if encoding:
                self.headers['Content-Encoding'] = encoding

        if self.media_type:
            if self.media_type.startswith('text/') and \
                    self.charset and 'charset' not in self.media_type:
                self.media_type += f'; charset={self.charset}'
            self.headers['Content-Type'] = self.media_type

        if self.download:
            self.headers['Content-Disposition'] = f'attachment; ' \
                                                  f'filename="{self.filename}"'
        if not self.last_modified:
            self.last_modified = time.strftime("%a, %d %b %Y %H:%M:%S GMT")
        self.headers['Last-Modified'] = self.last_modified

        if self.if_modified_since:
            self.if_modified_since = parse_date(
                self.if_modified_since.split(";")[0].strip())

        mtime = parse_date(self.last_modified)
        if self.if_modified_since is not None and mtime is not None and \
                self.if_modified_since >= int(mtime):
            self.headers['Date'] = time.strftime("%a, %d %b %Y %H:%M:%S GMT",
                                            time.gmtime())
        else:
            self.headers["Accept-Ranges"] = "bytes"
            if self.ranges:
                self.ranges = list(
                    parse_range_header(self.ranges, len(self.content)))
                if not self.ranges:
                    self.exc_info = "Requested Range Not Satisfiable"
                    self.status_code = 416
                offset, end = self.ranges[0]
                self.headers["Content-Range"] = "bytes %d-%d/%d" % (
                    offset, end - 1, len(self.content))
                self.headers["Content-Length"] = str(end - offset)
                self.content = file_iter_range(self.content, offset, end - offset)
                self.status_code = 206



from .components import Component
from toolkit.singleton import Singleton


class ServiceMeta(Singleton):

    def __new__(mcs, class_name, bases, props):
        """
        创建后更改return annotation
        :param class_name:
        :param bases:
        :param props:
        :return:
        """
        cls = super().__new__(mcs, class_name, bases, props)
        if "resolve" in props:
            props["resolve"].__annotations__['return'] = cls
        return cls


class Service(Component, metaclass=ServiceMeta):
    # 需要定义一下，不然会找到的父类的resolove。
    def resolve(self, *args, **kwargs):
        raise NotImplementedError()



import os
import sys
import asyncio

from IPython import embed
from IPython.core import formatters
from toolkit import cache_property

from apistar import Route
from apistar.http import PathParams, Response
from apistar.server.injector import ASyncInjector
from apistar.server.validation import VALIDATION_COMPONENTS
from apistar.server.asgi import ASGI_COMPONENTS, ASGIReceive,\
    ASGIScope, ASGISend

from ..solo.manager import MySelf
from ..bases.components import SettingsComponent
from ..helper import find_children, load_packages, get_real_method

# bug fix
formatters.get_real_method = get_real_method


class ConsoleManager(object):

    def __init__(self):
        sys.path.insert(0, os.getcwd())
        load_packages(".")
        SettingsComponent.register_path("settings")
        initial_components = {
            'scope': ASGIScope,
            'receive': ASGIReceive,
            'send': ASGISend,
            'exc': Exception,
            'app': ConsoleManager,
            'path_params': PathParams,
            'route': Route,
            'response': Response,
        }
        self.state = {
            'scope': MySelf(),
            'receive': MySelf(),
            'send': MySelf(),
            'exc': None,
            'app': self,
            'path_params': MySelf(),
            'route': MySelf()
        }
        self.injector = ASyncInjector(
            list(ASGI_COMPONENTS + VALIDATION_COMPONENTS) + find_children(),
            initial_components)

    async def resolve(self, type):
        def wrapper(arg: type):
            return arg

        return await self.injector.run_async([wrapper], dict(self.state))

    def __getattr__(self, item):
        assert item in self.beans, f"{item} cannot inject!"
        beans = self.beans[item]
        if len(beans) == 1:
            bean, _ = beans[0]
        else:
            i = input("Same bean name: {}: ".format(
                ", ".join(f"({index+1}) of {module}" for index, (_, module)
                                                in enumerate(beans))))
            bean = beans[int(i) - 1][0]
        return self.await(self.resolve(bean))

    def __getitem__(self, item):
        return self.__getattr__(item)

    @cache_property
    def beans(self):
        components = find_children(initialize=False)
        beans = dict()
        for component in components:
            type = component.resolve.__annotations__["return"]
            beans.setdefault(type.__name__, []).append((type, type.__module__))
        return beans

    @staticmethod
    def await(awaitable):
        loop = asyncio.get_event_loop()
        task = loop.create_task(awaitable.__await__())
        loop.run_until_complete(task)
        return task.result()

    def start(self):
        await = self.await
        embed()


def main():
    ConsoleManager().start()



import re
import uuid
import datetime

from apistar.exceptions import ValidationError

DATE_REGEX = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})$'
)

TIME_REGEX = re.compile(
    r'(?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
    r'(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?'
)

DATETIME_REGEX = re.compile(
    r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})'
    r'[T ](?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
    r'(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?'
    r'(?P<tzinfo>Z|[+-]\d{2}(?::?\d{2})?)?$'
)


class BaseFormat:
    def is_native_type(self, value):
        raise NotImplementedError()

    def validate(self, value):
        raise NotImplementedError()

    def to_string(self, value):
        raise NotImplementedError()


class DateFormat(BaseFormat):
    type = datetime.date

    def is_native_type(self, value):
        return isinstance(value, self.type)

    def validate(self, value):
        match = DATE_REGEX.match(value)
        if not match:
            raise ValidationError('Must be a valid date.')

        kwargs = {k: int(v) for k, v in match.groupdict().items()}
        return datetime.date(**kwargs)

    def to_string(self, value):
        try:
            return value.isoformat()
        except AttributeError:
            return str(value)


class TimeFormat(BaseFormat):
    type = datetime.time

    def is_native_type(self, value):
        return isinstance(value, self.type)

    def validate(self, value):
        match = TIME_REGEX.match(value)
        if not match:
            raise ValidationError('Must be a valid time.')

        kwargs = match.groupdict()
        kwargs['microsecond'] = kwargs['microsecond'] and kwargs['microsecond'].ljust(6, '0')
        kwargs = {k: int(v) for k, v in kwargs.items() if v is not None}
        return datetime.time(**kwargs)

    def to_string(self, value):
        try:
            return value.isoformat()
        except AttributeError:
            return str(value)


class DateTimeFormat(BaseFormat):
    type = datetime.datetime

    def is_native_type(self, value):
        return isinstance(value, self.type)

    def validate(self, value):
        match = DATETIME_REGEX.match(value)
        if not match:
            raise ValidationError('Must be a valid datetime.')

        kwargs = match.groupdict()
        kwargs['microsecond'] = kwargs['microsecond'] and kwargs['microsecond'].ljust(6, '0')
        tzinfo = kwargs.pop('tzinfo')
        if tzinfo == 'Z':
            tzinfo = datetime.timezone.utc
        elif tzinfo is not None:
            offset_mins = int(tzinfo[-2:]) if len(tzinfo) > 3 else 0
            offset_hours = int(tzinfo[1:3])
            delta = datetime.timedelta(hours=offset_hours, minutes=offset_mins)
            if tzinfo[0] == '-':
                delta = -delta
            tzinfo = datetime.timezone(delta)
        kwargs = {k: int(v) for k, v in kwargs.items() if v is not None}
        kwargs['tzinfo'] = tzinfo
        return datetime.datetime(**kwargs)

    def to_string(self, value):
        try:
            value = value.isoformat()
            if value.endswith('+00:00'):
                value = value[:-6] + 'Z'
            return value
        except AttributeError:
            return str(value)


class UUIDFormat(BaseFormat):
    type = uuid.UUID

    def is_native_type(self, value):
        return isinstance(value, self.type)

    def validate(self, value):
        return value

    def to_string(self, value):
        return str(value)


import typing
import inspect

from inspect import Parameter
from datetime import timedelta
from collections import namedtuple
from toolkit.singleton import Singleton
from toolkit.frozen import FrozenSettings
from toolkit.settings import SettingsLoader
from flask.sessions import SecureCookieSessionInterface
from apistar import Route, exceptions, http, Component as _Component

from .session import Session
from .controller import Controller


class Component(_Component, metaclass=Singleton):

    def identity(self, parameter: inspect.Parameter):
        """
        Each component needs a unique identifier string that we use for lookups
        from the `state` dictionary when we run the dependency injection.
        """
        parameter_name = parameter.name.lower()
        annotation_name = str(parameter.annotation)

        # If `resolve_parameter` includes `Parameter` then we use an identifier
        # that is additionally parameterized by the parameter name.
        args = inspect.signature(self.resolve).parameters.values()
        if inspect.Parameter in [arg.annotation for arg in args]:
            return annotation_name + ':' + parameter_name

        # Standard case is to use the class name, lowercased.
        return annotation_name

    def resolve(self, *args, **kwargs) -> object:
        raise NotImplementedError()

    def can_handle_parameter(self, parameter: inspect.Parameter):
        """重写这个方法是为了增加typing.Union类型的判定"""
        return_annotation = inspect.signature(self.resolve).return_annotation
        if return_annotation is inspect.Signature.empty:
            msg = (
                'Component "%s" must include a return annotation on the '
                '`resolve()` method, or override `can_handle_parameter`'
            )
            raise exceptions.ConfigurationError(msg % self.__class__.__name__)
        return type(return_annotation) == typing._Union and \
               parameter.annotation in return_annotation.__args__ or \
               parameter.annotation is return_annotation


class ServiceComponent(Component):
    """
    注入Service
    """
    def resolve(self, route: Route) -> Controller:
        return route.controller


class SettingsComponent(Component):
    """
    注入Settings
    """
    settings_path = None

    def __init__(self):
        self.settings = SettingsLoader().load(self.settings_path or "settings")

    def resolve(self) -> FrozenSettings:
        return self.settings

    @classmethod
    def register_path(cls, settings_path):
        cls.settings_path = settings_path


Cookie = typing.NewType('Cookie', str)


class CookiesComponent(Component):
    def resolve(self, cookie: http.Header) -> typing.Dict[str, Cookie]:
        cookies = dict()
        if cookie:
            for c in cookie.split(";"):
                key, val = c.strip().split("=", 1)
                cookies[key] = Cookie(val)
        return cookies


class CookieComponent(Component):

    def resolve(self,
                parameter: Parameter,
                cookies: typing.Dict[str, Cookie]) -> Cookie:
        return cookies.get(parameter.name.replace('_', '-'))


DummyFlaskApp = namedtuple(
    "DummyFlaskApp",
    "session_cookie_name,secret_key,permanent_session_lifetime, config")


class DummyFlaskAppComponent(Component):

    def __init__(self):
        self.default_config = {
            'SESSION_COOKIE_NAME': 'session',
            'SESSION_COOKIE_DOMAIN': None,
            'SESSION_COOKIE_PATH': '/',
            'SESSION_COOKIE_HTTPONLY': True,
            'SESSION_COOKIE_SECURE': False,
            'SESSION_COOKIE_SAMESITE': None,
            'SESSION_REFRESH_EACH_REQUEST': True,
            'SERVER_NAME': None,
        }

    def resolve(self, settings: FrozenSettings, host: http.Host) -> DummyFlaskApp:
                self.default_config["SERVER_NAME"] = host
                return DummyFlaskApp(
                    config=settings.get("SESSION_CONFIG", self.default_config),
                    secret_key=settings.get("SECRET_KEY"),
                    permanent_session_lifetime=timedelta(
                        days=settings.get("PERMANENT_SESSION_LIFETIME", 31)),
                    session_cookie_name=settings.get(
                        "SESSION_COOKIE_NAME", "session"))


class SessionComponent(Component):

    def __init__(self):
        self.dummy_request = namedtuple("Request", "cookies")
        self.session_interface = SecureCookieSessionInterface()

    def resolve(self,
                app: DummyFlaskApp,
                cookies: typing.Dict[str, Cookie]) -> Session:
            request = self.dummy_request(cookies=cookies)
            return self.session_interface.open_session(app, request)


# -*- coding:utf-8 -*-
from .app import Application, show_routes

from .build import main
from .build.tasks import Task

from .bases.service import Service
from .bases.session import Session
from .bases.hooks import Hook, Return
from .bases.controller import Controller
from .bases.response import FileResponse
from .bases.components import Component, Cookie

from .solo import Solo
from .solo.manager import SoloManager
from .console import main as console

from .types import Type, AsyncType, TypeEncoder, validators

from .route import route, get, post, delete, put, options

from .helper import redirect, require

__version__ = "0.7.3"
